/* 

   $Id: main.cc,v 1.29 2007/04/20 12:06:34 peekas Exp $

	Cadabra: an extendable open-source symbolic tensor algebra system.
	Copyright (C) 2001-2006  Kasper Peeters <kasper.peeters@aei.mpg.de>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, version 2.
	 
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	
*/

#include <string>
#include <map>
#include <iostream>
#include <modglue/main.hh>
#include <modglue/ext_process.hh>
#include <sstream>

#include "window.hh"

std::map<int, SigC::Connection> connections;

std::ofstream  debugout;

bool verify_breqn_presence()
	{
	char templ[10]="cdbXXXXXX";
	std::ostringstream total;

	// First test LaTeX without any style files.
	int fd = mkstemp(templ);
	total << "\\documentclass{article}\n"
			<< "\\begin{document}\n"
			<< "test\n"
			<< "\\end{document}\n";
	write(fd, total.str().c_str(), total.str().size());
	close(fd);
	std::string nf=std::string(templ)+".tex";
	rename(templ, nf.c_str());
	std::string cmd=std::string("latex --interaction nonstopmode ")+nf+" >/dev/null 2>&1";
	int ret=system(cmd.c_str());
	unlink(nf.c_str());	
	cmd=std::string(templ)+".aux";
	unlink(cmd.c_str());
	cmd=std::string(templ)+".dvi";
	unlink(cmd.c_str());
	cmd=std::string(templ)+".log";
	unlink(cmd.c_str());
	if(ret==-1) {
		std::cerr << "XCadabra: LaTeX not installed." << std::endl;
		sleep(1);
		return false;
		}
	if(ret!=0) {
		std::cerr << "XCadabra: LaTeX not installed properly." << std::endl;
		sleep(1);
		return false;
		}

	// Then test whether breqn can be included.
	char templ2[10]="cdbXXXXXX";
	fd = mkstemp(templ2);
	total.str("");
	total << "\\documentclass{article}\n"
			<< "\\usepackage{breqn}\n"
			<< "\\begin{document}\n"
			<< "test\n"
			<< "\\end{document}\n";
	write(fd, total.str().c_str(), total.str().size());
	close(fd);
	nf=std::string(templ2)+".tex";
	rename(templ2, nf.c_str());
	cmd=std::string("latex --interaction nonstopmode ")+nf+" >/dev/null 2>&1";
	ret=system(cmd.c_str());
	unlink(nf.c_str());	
	cmd=std::string(templ2)+".aux";
	unlink(cmd.c_str());
	cmd=std::string(templ2)+".dvi";
	unlink(cmd.c_str());
	cmd=std::string(templ2)+".log";
	unlink(cmd.c_str());
	if(ret==-1) {
		std::cerr << "XCadabra: LaTeX not installed." << std::endl;
		sleep(1);
		return false;
		}
	if(ret!=0) {
		std::cerr << "XCadabra: the breqn.sty style file is not in LaTeX's TEXINPUTS." << std::endl;
		sleep(1);
		return false;
		}
	
	return true;
	}


int main (int argc, char *argv[])
	{
	// Open the debug output file, if requested.
	char hostname[256];
	gethostname(hostname, 255);
	char *pbs_job=getenv("PBS_JOBID");
	std::string logname=std::string("cdb_")+hostname+(pbs_job==0?"":std::string("_")+pbs_job)
                       +std::string(".log");

	char *cdblog=getenv("CDB_LOG");
	if(cdblog==0)
		debugout.open("/dev/null", std::ios::app);
	else switch(atoi(cdblog)) {
		case 1:
			debugout.open(logname.c_str(), std::ios::app);
			break;
		case 2:
			debugout.open("/dev/tty", std::ios::app);
			break;
		}

	// Ensure correct installation.
	if(!verify_breqn_presence())
		return -1;

	Gtk::Main     kit(&argc, &argv);
	modglue::main mm(argc, argv);

	// Set the environment for cadabra.
//	setenv("CDB_USE_UTF8", "1", 1);
	unsetenv("CDB_PRINTSTAR");//, "false", 1);

	// Argument parsing
	std::string filename;
	if(argc>1)
		filename=argv[1];

//	modglue::ext_process ls_proc(std::string(DESTDIR)+std::string("/bin/cadabra"));
	modglue::ext_process ls_proc("cadabra");
	ls_proc << "--xcadabra";
	ls_proc.setup_pipes(); // FIXME: need cleaner error messages if this is forgotten
	mm.add(&ls_proc);
	XCadabra theiface(ls_proc, filename, &mm);

	ls_proc.input_pipe("stdout")->receiver.connect(sigc::mem_fun(theiface, &XCadabra::receive));
	ls_proc.input_pipe("stderr")->receiver.connect(sigc::mem_fun(theiface, &XCadabra::receive));
	mm.process_died.connect(sigc::mem_fun(theiface, &XCadabra::on_kernel_exit));

	// FIXME: if we run this afterwards, the pipes are not setup yet. Note: this also
	// means that if we want to start a program while in the main loop, we have to
	// do this connecting business again. For the time being, let cadabra run forever
	// and don't worry too much about this.

	ls_proc.fork(); // FIXME: need error message upon failure
	*(ls_proc.output_pipe("stdin")) << "@print_status{true};\n" << std::flush;

	theiface.connect_io_signals();

	// If done this way, we miss signals if the process exits before
	// the main loop is entered. Well, cadabra doesn't do that.

//	mm.run(0);
	Gtk::Main::run();
	return 0;
	} 
