
#include "props.hh"
#include "modules/dummies.hh"
#include <iostream>

properties::property_map_t            properties::props;
properties::pattern_map_t             properties::pats;
properties::registered_property_map_t properties::registered_properties;

void properties::register_properties()
	{
	register_property(&create_property<IndexInherit>);
	register_property(&create_property<PropertyInherit>);
	}

pattern::pattern()
	{
	}

pattern::pattern(const exptree::iterator_base& it)
	: obj(it), headnode(it->name)
	{
	}

bool pattern::match(const exptree::iterator& it) const
	{
	// Special case for range wildcards.
	if(it->name==obj.begin()->name && children_wildcard()) {
		exptree::iterator hm=obj.begin(obj.begin());
		if(exptree::number_of_children(hm)==0) {
			return true; // # without arguments
			}
		exptree::iterator hmarg=hm.begin();
		exptree::iterator seqarg=hm; 
		const Indices *ind=0;

		if(*hmarg->name=="\\comma" || *hmarg->name!="\\sequence") {
			exptree::iterator stt=hmarg;
			if(*hmarg->name=="\\comma") {
				stt=hmarg.begin();
				seqarg=hmarg.begin();
				seqarg.skip_children();
				++seqarg;
				}
			ind=properties::get<Indices>(stt);
			}
		else seqarg=hmarg;

		if(seqarg!=hm) {
			exptree::sibling_iterator seqit=seqarg.begin();
			unsigned int from=to_long(*seqit->multiplier);
			++seqit;
			unsigned int to  =to_long(*seqit->multiplier);

			if(exptree::number_of_children(it)<from ||
				exptree::number_of_children(it)>to ) 
				return false;
			}
		
		if(ind!=0) {
			exptree::sibling_iterator indit=it.begin();
			while(indit!=it.end()) {
				const Indices *gi=properties::get<Indices>(indit);
				if(gi!=ind) {
					return false;
					}
				++indit;
				}
			}
		
		return true;
		}

	// Cases without range wildcard.
//	txtout << "comparing " << *obj.begin()->name << " " << *it->name << std::endl;
	int res=subtree_compare(it, obj.begin(), 0, true, 0);
	if(abs(res)<=1) return true;

	return false;
	}

bool pattern::children_wildcard() const	
	{
	if(exptree::number_of_children(obj.begin())==1) 
		if(obj.begin(obj.begin())->is_range_wildcard())
			return true;
	return false;
	}

bool properties::has(const property_base *pb, exptree::iterator it) 
	{
	std::pair<property_map_t::iterator, property_map_t::iterator> pit=props.equal_range(it->name);
	while(pit.first!=pit.second) {
//		txtout << typeid(pit.first->second.second).name() << " versus " 
//				 << typeid(pb).name() << std::endl;
		if(typeid(pit.first->second.second)==typeid(pb) && pit.first->second.first->match(it))  // match found
			return true;
		++pit.first;
		}
	return false;
	}

void properties::clear() 
	{
	props.clear();
	}

void properties::register_property(property_base* (*fun)())
	{
	property_base *tmp=fun(); // need a property object of this type temporarily to retrieve the name
	registered_properties[tmp->name()]=fun;
	delete tmp; 
	}

bool property_base::parse(exptree& tr, exptree::iterator, exptree::iterator arg, keyval_t& keyvals)
	{
	if(tr.number_of_children(arg)==0) return true;
//	txtout << name() << ": should not have any arguments." << std::endl;
	return false;
	}

bool property_base::parse_one_argument(exptree::iterator arg, keyval_t& keyvals)
	{
	if(*arg->name=="\\equals") {
		exptree::sibling_iterator key=arg.begin();
		if(key==arg.end()) return false;
		exptree::sibling_iterator val=key;
		++val;
		if(val==arg.end()) return false;
		keyvals.insert(keyval_t::value_type(*arg.begin()->name, val));
		}
	else {
		if(unnamed_argument()!="") {
			keyvals.insert(keyval_t::value_type(unnamed_argument(), arg));
			}
		else return false;
		}
	return true;
	}

bool property_base::preparse_arguments(exptree::iterator prop, keyval_t& keyvals) 
	{
	if(exptree::number_of_children(prop)==0) return true;
	if(exptree::number_of_children(prop)>1) return false;
	if(*prop.begin()->name!="\\comma") { // one argument
		if(parse_one_argument(prop.begin(), keyvals)==false)
			return false;
		}
	else {
		exptree::sibling_iterator sib=prop.begin().begin();
		while(sib!=prop.begin().end()) {
			if(parse_one_argument(sib, keyvals)==false)
				return false;
			++sib;
			}
		}
	return true;
	}


void property_base::display(std::ostream& str) const
	{ 
	str << name();
	}

std::string property_base::unnamed_argument() const
	{
	return "";
	}

bool property_base::core_parse(keyval_t& keyvals)
	{
	return true;
	}

bool labelled_property::core_parse(keyval_t& keyvals)
	{
	property_base::keyval_t::iterator lit=keyvals.find("label");
	if(lit!=keyvals.end()) {
		label=*lit->second->name;
		return true;
		}
	else {
//		txtout << "This property needs a label." << std::endl;
		return false;
		}
	}
	
bool operator<(const pattern& one, const pattern& two)
	{
	return tree_less(one.obj, two.obj);
//	if(*(one.obj.begin()->name)<*(two.obj.begin()->name)) return true;
//	return false;
	}

//bool operator==(const pattern& one, const pattern& two)
//	  {
//	  return one.obj==two.obj; // FIXME: handle dummy indices
//	  }

void properties::insert_prop(exptree::iterator it, const property *pr)
	{
//	assert(pats.find(pr)==pats.end()); // identical properties have to be assigned through insert_list_prop

	pattern *pat=new pattern(it);

	std::pair<property_map_t::iterator, property_map_t::iterator> pit=
		props.equal_range(it->name);

	property_map_t::iterator first_nonpattern=pit.first;

	while(pit.first!=pit.second) {
		// keep track of the first non-pattern element
		if(exptree::number_of_children((*pit.first).second.first->obj.begin())==1) 
			if((*pit.first).second.first->obj.begin().begin()->is_range_wildcard()) 
				++first_nonpattern;
			
		if((*pit.first).second.first->match(it)) { // match found
			if(typeid(*pr)==typeid(*(*pit.first).second.second)) {
				const labelled_property *lp   =dynamic_cast<const labelled_property *>(pr);
				const labelled_property *lpold=dynamic_cast<const labelled_property *>(pit.first->second.second);
				if(!lp || !lpold || lp->label==lpold->label) {
//					txtout << "Removing previously set property." << std::endl;
					pattern  *oldpat=pit.first->second.first;
					const property_base *oldprop=pit.first->second.second;
					props.erase(pit.first);
					pats.erase(oldprop);
					delete oldpat;
					delete oldprop;
					break;
					}
				}
			}
		++pit.first;
		}

	pats.insert(pattern_map_t::value_type(pr, pat));
	properties::props.insert(property_map_t::value_type(it->name, pat_prop_pair_t(pat,pr)));
	}

void properties::insert_list_prop(const std::vector<exptree::iterator>& its, const list_property *pr)
	{
	assert(pats.find(pr)==pats.end()); // identical properties have to be assigned through insert_list_prop
	assert(its.size()>0);

	for(unsigned int i=0; i<its.size(); ++i) {
		pattern *pat=new pattern(its[i]);
		
		std::pair<property_map_t::iterator, property_map_t::iterator> pit=
			props.equal_range(its[i]->name);
		while(pit.first!=pit.second) {
			if((*pit.first).second.first->match(its[i])) { // match found
				if(typeid(*pr)==typeid(*(*pit.first).second.second)) {
//					txtout << "Removing previously set list property." << std::endl;
					// FIXME: FIX THIS!!!!
//					pattern  *oldpat=pit.first->second.first;
//					const property_base *oldprop=pit.first->second.second;
//					props.erase(pit.first);
//					pats.erase(oldprop);
//					delete oldpat;
//					delete oldprop;
					break;
					}
				}
			++pit.first;
			}
		
		pats.insert(pattern_map_t::value_type(pr, pat));
		properties::props.insert(property_map_t::value_type(its[i]->name, pat_prop_pair_t(pat,pr)));
		}
//		insert_prop(its[i], pr);
	}

//properties::property_map_t::iterator properties::get_equivalent(exptree::iterator it,
//																					 property_map_t::iterator props)
//	{
//	// first collect list of properties of the original
//
//	// then find a new property with the same 
//	return props;
//	}
